/******************************************************************************/
/***                                                                        ***/
/***  Name : agent.c                                                        ***/
/***                                                                        ***/
/******************************************************************************/
/***                                                                        ***/
/***  Description :                                                         ***/
/***     This file contains a "main" segment for using the core agent.      ***/
/***     The main  segment initializes the core agent through a call to     ***/
/***     the ca_init() function.                                            ***/
/***     The program then waits for incoming messages and calls the core    ***/
/***     agent for processing these messages.                               ***/
/***                                                                        ***/
/******************************************************************************/

#define __EXTENSIONS__ /* Because getopt() is used later */

#include <assert.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#ifdef WIN32
#include <xom.h>
#include <xmp.h>
#include <usr_caDefs.h>
#include <usr_caGo.h>
#include <usr_caNoti.h>
#include <usr_caDmi.h>
#include <usr_caImp.h>
#include <usr_caCtrl.h>
#include <usr_caXom.h>
#else /* WIN32 */
#include <cmip/xom.h>
#include <cmip/xmp.h>
#include <tmn_agent/usr_caDefs.h>
#include <tmn_agent/usr_caGo.h>
#include <tmn_agent/usr_caNoti.h>
#include <tmn_agent/usr_caDmi.h>
#include <tmn_agent/usr_caImp.h>
#include <tmn_agent/usr_caCtrl.h>
#include <tmn_agent/usr_caXom.h>
#endif /* WIN32 */
#include "metadata.h"
#include "init.h"

/*
** External declarations
*/
extern int getopt(int argc, char * const *argv, const char *optstring) ;
extern char * optarg ;


#define USAGE "usage : %s [-N nsap] [-T tsel] [-S ssel] [-P psel] [-x trace-level]"

/*
** Global context for the core agent
*/
static ca_context_t context = {
	NULL,
	MP_DEFAULT_SESSION,
	MP_DEFAULT_CONTEXT,
	stdout,
	CA_L_NO_TRACE,
	10,
	OM_TRUE,
	OM_TRUE,
	10,
	NULL
} ;


/*
** Session template.
** It is used to build the session included in the context.
*/ 

#define POS_OF_N_ADDRESSES 	1
#define POS_OF_P_SELECTOR 	2
#define POS_OF_S_SELECTOR 	3
#define POS_OF_T_SELECTOR 	4

OM_IMPORT(MP_C_PRESENTATION_ADDRESS)
OM_IMPORT(MP_C_SESSION)
static OM_descriptor   agent_addr[] = {
	OM_OID_DESC(OM_CLASS, MP_C_PRESENTATION_ADDRESS),
	{MP_N_ADDRESSES, OM_S_OCTET_STRING, {0, 0}},
	{MP_P_SELECTOR, OM_S_OCTET_STRING, {OM_STRING("rfc1")}},
	{MP_S_SELECTOR, OM_S_OCTET_STRING, {OM_STRING("Prs")}},
	{MP_T_SELECTOR, OM_S_OCTET_STRING, {OM_STRING("CMIP")}},
	OM_NULL_DESCRIPTOR
};
static OM_descriptor   session_addr[] = {
	OM_OID_DESC(OM_CLASS, MP_C_SESSION),
	{MP_REQUESTOR_ADDRESS, OM_S_OBJECT,  { 0, agent_addr}},
	{MP_FILE_DESCRIPTOR, OM_S_INTEGER, 0},
	{MP_ROLE, OM_S_INTEGER, MP_T_PERFORMING | MP_T_REPORTING},
	OM_NULL_DESCRIPTOR
};


/*
** Prototypes
*/

static int initContext(int inArgc, char * inArgv[]) ;
static int initAddress(int pos, char * value) ;
static void killproc(int sig) ;
static int hextoint(char c) ;


/*****************************************************************************
** FUNCTION:   main
**----------------------------------------------------------------------------
**
**  Description:
**  ------------
**   This is the main function of the agent.
**   It parses the command line options and sets up the toolkit.
** 
*****************************************************************************/

int main(int argc, char * argv[])
{
	int result = 0 ;
	OM_return_code omrc ;


	/*
	** Declare callback for signal 
	*/
	signal(SIGINT, killproc);

	result = initContext(argc, argv) ;

	/*
	** Initialize the DMI library. 
	** This needs to be done even if the only functions
	** used in the DMI library are the utility     
	** routines, such as ca_dmi_add_package.        
	** Note also that the initialization should occur 
	** before ca_init() calls the init() callbacks of
	** the classes in listOfClass.      
	*/
	if (result == 0) 
		ca_dmi_init(&context);

	/*
	** Initialize the core agent
	*/
	if (result == 0) {
		omrc = ca_init(listOfClass, &context) ;
		if (omrc != OM_SUCCESS) {
			printf("ca_init() returned %d\n", omrc) ;
			result = 1 ;
		}
		else if (context.trace) {
			printf("CA: context.session = {\n");
			OMPexamin(context.session, 2, 0) ;
			printf("CA: }\n") ;
		}
	}

	/*
	** Initialize the managed object classes.
	** Each managed object class has an initialization
	** function which is called by general_init().
	** The object classes are initialized in the order
	** in which their initialization functions are called.
	*/
	if (result == 0) {
		omrc = general_init(&context) ;
		if (omrc != OM_SUCCESS) {
			printf("general_init() returned %d\n", omrc) ;
			result = 1 ;
		}
	}

	/*
	** Include moc and moi in single replies.
	*/
	if (result == 0) {
		omrc = ca_addMocAndMoi(&context, OM_TRUE) ;
		if (omrc != OM_SUCCESS) {
			printf("ca_addMocAndMoi() returned %d\n", omrc) ;
			result = 1 ;
		}
	}

	/*
	** The ca_callback() function waits for incoming messages using the
	** the mp_wait() function of the XMP API. The length of the time
	** ca_callback() waits corresponds to the value provided in the
	** context when calling the ca_init() function.  When developing
	** an agent which uses the "select" mechanism for waiting for different
	** events, the ca_callback() function can be associated with the file
	** descriptor contained in the session member of the context 
	** object. This file descriptor can be retreived by calling :
	**		ca_xom_getSessionFd(context.session, &fd)
	** 
	*/

	if (result == 0) {
		signal(SIGINT, killproc);
		while (1) {
			ca_callback(&context) ;
			ca_dmi_tick(&context) ;
		}
	}

	return result ;
}



/*****************************************************************************
** FUNCTION:   initContext
**----------------------------------------------------------------------------
**
**  Description:
**  ------------
**   This function initializes the toolkit context according the options
**   specified on the command line.
** 
*****************************************************************************/

static int initContext(int inArgc, char * inArgv[])
{
	int result = 0 ;
	OM_return_code omrc ;
	int c ;

	/*
	** Parse the command line.
	** 
	*/
	c = getopt(inArgc, inArgv, "N:P:S:T:x:") ;
	while ((c != EOF) && (result == 0)) {
		switch(c) {
			case 'x' :
				/*
				** Update the trace level to be used by 
				** the core agent. 
				*/
				context.trace = atoi(optarg);
				break ;
			case 'N' :
				result = initAddress(POS_OF_N_ADDRESSES,optarg);
				if (result != 0)
					(void)fprintf(stderr, 
						"%s is invalid\n", optarg) ;
				break ;
			case 'P' :
				result = initAddress(POS_OF_P_SELECTOR, optarg);
				if (result != 0)
					(void)fprintf(stderr, 
						"%s is invalid\n", optarg) ;
				break ;
			case 'S' :
				result = initAddress(POS_OF_S_SELECTOR, optarg);
				if (result != 0)
					(void)fprintf(stderr, 
						"%s is invalid\n", optarg) ;
				break ;
			case 'T' :
				result = initAddress(POS_OF_T_SELECTOR, optarg);
				if (result != 0)
					(void)fprintf(stderr, 
						"%s is invalid\n", optarg) ;
				break ;
			default :
				(void)fprintf(stderr, USAGE, inArgv[0]) ;
				result = 1 ;
				break ;
		}
		c = getopt(inArgc, inArgv, "N:P:S:T:x:") ;
	}

	/*
	** Create the workspace.
	*/
	if (result == 0) {
		context.wks = mp_initialize() ;
		if (context.wks == NULL) {
			printf("CA: INITIALISATION OF XMP API FAILED\n");
			result = 1 ;
		}
	}

	/*
	** Build a session which will contain the local address used       
	** to bind the agent.
	*/
	if (result == 0) {
		omrc = om_create(MP_C_SESSION, OM_FALSE, context.wks,
					&context.session) ;
		if (omrc == OM_SUCCESS) 
			omrc = om_put(context.session, OM_REPLACE_ALL,
					session_addr, NULL, NULL, NULL) ;
		if (omrc != OM_SUCCESS) {
			printf("CA: FAILED TO BUILD SESSION\n");
			result = 1 ;
		}
		else if (context.trace) {
			printf("CA: SESSION TO USE:\n");
			OMPexamin(context.session, 2, 0);
		}

	}

	/*
	** Update default values:
	** ---------------------
	**   checkClass: when set to true, it specifies that the core agent
	**  	     must check that the base managed object class specified
	**           in an incoming message is actually managed by the agent;
	**           that is, contained in the "listOfClass" structure given
	**           at start-up time.                                         
	**   timeout: specifies the length of time ca_callback() waits for
	**           an incoming message. It is the timer used by the core
	**           agent when calling the mp_wait() function.
        **   maxOutstanding: indicates the maximum number of outstanding     
	**           operations the core agent should accept. This parameter
	**           is only relevant when using the LL API in asynchronous 
	**           mode.                                                  
	*/

	context.checkClass = OM_TRUE;
	context.timeout = 10;
	context.maxOutstanding = 10;

	return result ;
}


/*****************************************************************************
** FUNCTION:   initAddress
**----------------------------------------------------------------------------
**
**  Description:
**  ------------
**   This function assigns a value to a specific field in the address. 
** 
*****************************************************************************/

static int initAddress(int pos, char * value)
{
	if (strncmp(value, "0x", 2) != 0) {
		agent_addr[pos].value.string.elements = value ;
		agent_addr[pos].value.string.length = strlen(value) ;
	}
	else {
		int byteCount, i ;
		char left, right ;
		/*
		** Translate value if it is in hexadecimal form
		*/
		byteCount = (int)(strlen(value) - 2) / 2 ;
		for (i = 0 ; i < byteCount ; i++) {
			left = value[2 * i + 2] ;
			right = value[2 * i + 3] ;
			value[i] = hextoint(left) * 16 + hextoint(right) ;
		}

		agent_addr[pos].value.string.elements = value ;
		agent_addr[pos].value.string.length = byteCount ;		
	}

	return 0 ;
}

static int hextoint(char c)
{
   if (isdigit(c))
      return (0xFF & (c - '0'));
   return (10 + (0xFF & (toupper(c) - 'A')));
}



/*****************************************************************************
** FUNCTION:   killproc
**----------------------------------------------------------------------------
**
**  Description:
**  ------------
**   Interrupt handler. This function performs a clean exit. 
** 
*****************************************************************************/

static void killproc(int sig)
{
	MP_status       status;

	printf("\nCA: RELEASE SESSION\n");
	status = mp_unbind(context.session);
	if (status != MP_SUCCESS) {
		printf("\tfails to release session\n");
	}
	printf("CA: SHUTDOWN THE API\n");
	mp_shutdown(context.wks);
	exit(0);
}


