/******************************************************************************/
/* logDB Library Code                                                         */
/* ==================                                                         */
/*                                                                            */
/* This code is provided for example purposes in order to show how the        */
/* Database API works. Normally this code will be replaced by user supplied   */
/* code which will better fit the users requirements and the method of        */
/* storage best suited to the application.                                    */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/* How it Works                                                               */
/* ============                                                               */
/*                                                                            */
/* The log is stored physically in a file in /var/tmp.  This file is created  */
/* whenever an instance of a log object is created, and is deleted whenever   */
/* that instance is deleted.                                                  */
/*                                                                            */
/* If the action to save the log is received by an instance of a log object   */
/* then a backup file (with a .sav suffix) and an index file (.index suffix)  */
/* file are created.  The backup file contains the contents of the log.  The  */
/* index file contains information about the log itself and a list of offsets */
/* to log records in the backup file. This means that the contents of the log */
/* can be restored if necessary.                                              */
/*                                                                            */
/* A small amount of information about each log record (size, id etc.) is     */
/* stored in memory to provide a quick look-up capability.  This is 16 bytes  */
/* per log record.  This improves performance, by not having to scan the file */
/* each time to retrieve this information. It is this information which is    */
/* stored in the index file whenever the log is saved.                        */
/*                                                                            */
/* Information about the log itself is also stored in a block of memory. The  */
/* most important information stored here are a pair of integers which store  */
/* offsets in the file. One is the "write offset" which is the position in the*/
/* file where the next log record is to be added.  The other is the "delete   */
/* offset where the next log record to be deleted starts.                     */
/*                                                                            */
/* For a non-wrapping log, the write offset is always the end of the file,    */
/* and the delete offset is not used.  For a wrapping log, the write offset   */
/* may point to a part of the file which is already written and results in    */
/* the previous information being overwritten.  The delete offset is used as  */
/* a check in this case to ensure that valid information is not being         */
/* overwritten.                                                               */
/*                                                                            */
/* All functions are implemented as required for both a linear log (halt mode)*/
/* or a wrapped log (wrap mode).                                              */
/*                                                                            */
/******************************************************************************/

/*
 * Standard include files.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <time.h>
#include <sys/param.h>

/*
 * XOM/XMP include files.
 */
#include <cmip/xom.h>
#include <cmip/xmp.h>
#include <cmip/xmp_cmis.h>

/*
 * Agent Toolkit include files.
 */
#include <tmn_agent/usr_caDefs.h>
#include <tmn_agent/usr_caGo.h>
#include <tmn_agent/usr_caDmi.h>
#include <tmn_agent/usr_caXom.h>
#include <tmn_agent/usr_caCtrl.h>
#include <tmn_agent/usr_caTree.h>
#include <tmn_agent/usr_caNoti.h>
#include <tmn_agent/usr_caLog.h>

/*
 * Local include files.
 */
#include <logDB.h>

/******************************************************************************/
/*                           UTILITY FUNCTIONS                                */
/******************************************************************************/

/******************************************************************************/
/* Function: printInline                                                      */
/*                                                                            */
/* This function prints out the contents of an inline log record in a human   */
/* readable format.  It is used for tracing purposes.                         */
/*                                                                            */
/******************************************************************************/

static void
printInline(
	ca_dmi_logR_inline_t	*rec
	) {

	char 		*ptr = (char *) rec->rec;
	OM_descriptor	tmp;
	OM_uint32	tmpi;
	int		i, j, nrdn;

	/*
	 * Record size.
	 */
	memcpy(&tmpi, ptr, sizeof(OM_uint32));
	printf("Log record size\t\t%d\n", tmpi);
	ptr += sizeof(OM_uint32);

	/*
	 * Record Id.
	 */
	memcpy(&tmpi, ptr, sizeof(OM_uint32));
	printf("Log record Id\t\t%d\n", tmpi);
	ptr += sizeof(OM_uint32);

	/*
	 * Logging Time.
	 */
	memcpy(&tmp, ptr, sizeof(OM_descriptor));
	ptr += sizeof(OM_descriptor);
	printf("Logging time\t\t");
	for (i = 0; i < tmp.value.string.length; i++) {
		printf("%c", ptr[i] & 0xff);
	}
	printf("\n");
	ptr += tmp.value.string.length;

	/*
	 * MOC
	 */
	memcpy(&tmp, ptr, sizeof(OM_descriptor));
	ptr += sizeof(OM_descriptor);
	printf("MOC\t\t\t0x");
	for (i = 0; i < tmp.value.string.length; i++) {
		printf("%02x", ptr[i] & 0xff);
	}
	printf("\n");
	ptr += tmp.value.string.length;

	/*
	 * Number of RDNs.
	 */
	memcpy(&tmpi, ptr, sizeof(OM_uint32));
	nrdn = tmpi;
	printf("Number of RDNs\t\t%d\n", nrdn);
	ptr += sizeof(OM_uint32);
	

	/*
	 * RDNs
	 */
	printf("RDNs:");
	for (i = 0; i < nrdn; i++) {

		/*
		 * ID
		 */
		memcpy(&tmp, ptr, sizeof(OM_descriptor));
		ptr += sizeof(OM_descriptor);
		printf("\t\t\t0x");
		for (j = 0; j < tmp.value.string.length; j++) {
			printf("%02x", ptr[j] & 0xff);
		}
		ptr += tmp.value.string.length;
		printf(" = ");

		memcpy(&tmp, ptr, sizeof(OM_descriptor));
		ptr += sizeof(OM_descriptor);
		if (!(
			(tmp.syntax & OM_S_SYNTAX) == OM_S_NO_MORE_SYNTAXES ||
			(tmp.syntax & OM_S_SYNTAX) == OM_S_BOOLEAN ||
			(tmp.syntax & OM_S_SYNTAX) == OM_S_ENUMERATION ||
			(tmp.syntax & OM_S_SYNTAX) == OM_S_INTEGER ||
			(tmp.syntax & OM_S_SYNTAX) == OM_S_NULL ||
			(tmp.syntax & OM_S_SYNTAX) == OM_S_OBJECT
			)) {
			printf("0x");
			for (j = 0; j < tmp.value.string.length; j++) {
				printf("%02x", ptr[j] & 0xff);
			}
			ptr += tmp.value.string.length;
		}
		else {
			printf("%d", tmp.value.integer);
		}
		printf("\n");
	}

	/*
	 * Event type.
	 */
	memcpy(&tmp, ptr, sizeof(OM_descriptor));
	ptr += sizeof(OM_descriptor);
	printf("Event type\t\t0x");
	for (i = 0; i < tmp.value.string.length; i++) {
		printf("%02x", ptr[i] & 0xff);
        }
	printf("\n");
	ptr += tmp.value.string.length;

	/*
	 * Notification info.
	 */
	memcpy(&tmp, ptr, sizeof(OM_descriptor));
	ptr += sizeof(OM_descriptor);
	printf("Info:\n\t");
	for (i = 0; i < tmp.value.string.length; i++) {
		printf("%02x ", ptr[i] & 0xff);
		if ((i % 20) == 19) {
			printf("\n\t");
		}
	}
	printf("\n");
	
	return;
}

/******************************************************************************/
/* Function: free_logdb_rec                                                   */
/*                                                                            */
/* Used to free an entire linked list of the 16 byte information blocks which */
/* store critical information about a log record stored in the log.           */
/*                                                                            */
/******************************************************************************/

static void
free_logdb_rec(
	logdb_file	*log
	) {

	logdb_rec	*curr;
	logdb_rec	*next;

	curr = log->rec;
	while (curr != NULL) {
		next = curr->next;
		free(curr);
		curr = next;
	}

	log->rec = NULL;
	log->last = NULL;
}

/******************************************************************************/
/* Function: free_logdb_file                                                  */
/*                                                                            */
/* Frees block containing local information about the log itself. This        */
/* includes freeing the list of information blocks for individual log records.*/
/*                                                                            */
/******************************************************************************/

static void
free_logdb_file(
	logdb_file	*log
	) {

	free_logdb_rec(log);
	free(log->fileName);
	free(log);

	return;
}

/******************************************************************************/
/* Function: find_logdb_rec                                                   *//*                                                                            */
/* Find the information block for a log record with a supplied log record ID. */
/*                                                                            */
/******************************************************************************/

static logdb_rec *
find_logdb_rec(
	logdb_file	*log,
	OM_uint32	id
	) {

	logdb_rec	*curr;

	curr = log->rec;
	while(curr != NULL) {
		if (curr->id == id) {
			break;
		}
		curr = curr->next;
	}

	return(curr);
}

/******************************************************************************/
/* Function: copy_file                                                        */
/*                                                                            */
/* Copies the contents of the file with filename pointed to by "src" to the   */
/* file with the filename pointed to by "dst".                                */
/*                                                                            */
/******************************************************************************/

static int
copy_file(const char *src, const char *dst)
{
	FILE	*srcfd;
	FILE	*dstfd;
	char	buffer[BUFSIZ];
	int	c;

	srcfd = fopen(src, "r");
	dstfd = fopen(dst, "w");
	if ((srcfd == NULL) || (dstfd == NULL)) {
		if (srcfd != NULL) fclose(srcfd);
		if (dstfd != NULL) fclose(dstfd);
		return(1);
	}

	do {
		c = fread(buffer, 1, sizeof(buffer), srcfd);
		if (c > 0) {
			c = fwrite(buffer, 1, c, dstfd);
		}
	}
	while (!(feof(srcfd) || ferror(srcfd) || ferror(dstfd)));

	if (ferror(dstfd) || ferror(srcfd)) {
		fclose(srcfd);
		fclose(dstfd);
		return(1);
	}
	else {
		fclose(srcfd);
		return(fclose(dstfd));
	}	
}

/******************************************************************************/
/* Function: delete_logdb_rec                                                 */
/*                                                                            */
/* Delete and free a log record information block.                            */
/*                                                                            */
/******************************************************************************/

static void
delete_logdb_rec(
	logdb_file	*log,
	OM_uint32	id
	) {

	logdb_rec	*curr;
	logdb_rec	*prev;

	curr = log->rec;
	prev = NULL;
	while (curr != NULL) {
		if (curr->id == id) {
			if (prev == NULL) {
				log->rec = curr->next;
			}
			else {
				prev->next = curr->next;
			}
			free(curr);
			break;
		}
		prev = curr;
		curr = curr->next;
	}

	return;
}

/******************************************************************************/
/*                               API FUNCTIONS                                */
/******************************************************************************/

/******************************************************************************/
/* Function: logDB_create                                                     */
/*                                                                            */
/* Create the physical log.  Here we do the following:                        */
/* 1. allocate some memory to hold information about the physical log.        */
/* 2. derive a name for the file used implement the physical log. This is     */
/*	derived from the logId.                                               */
/* 3. create and open the file.                                               */
/* Note that the handle returned is actually the pointer to the block of      */
/* memory which holds information about the log.                              */
/*                                                                            */
/******************************************************************************/

logDB_handle
logDB_create(
	ca_context_t	*cont,
	OM_descriptor	*logId,
	OM_uint32	logWrap,
	OM_uint32	maxLogSize
	) {

	logdb_file	*log;
	char		filename[MAXPATHLEN];

	/*
	 * Tracing.
	 */
	if (cont->trace) {
		printf("logDB_create: start, logWrap = %d, maxLogSize = %d\n",
					logWrap, maxLogSize);
	}

	/*
	 * Allocate some memory for the local log structure.
	 */
	log = (logdb_file *) malloc(sizeof(logdb_file));
	if (log == NULL) {
		if (cont->trace) {
			printf("logDB_create: out of memory.\n");
		}
		return(NULL);
	}

	/*
	 * Initialize parameters in local log structure.
	 */
	log->logWrap = logWrap;
	log->maxLogSize = maxLogSize;
	log->woff = 0;
	log->rec = NULL;
	log->last = NULL;

	/*
	 * Log Id is a descriptor which can contain an integer or a Graphic
	 * String.  Generate a filename based on the Log Id. The file is created
	 * in /var/tmp.
	 */
	switch (logId->syntax) {

	case OM_S_INTEGER:

		sprintf(filename, "/var/tmp/agentlog_%d", logId->value.integer);
		break;

	case OM_S_GRAPHIC_STRING:

		sprintf(filename, "/var/tmp/agentlog_%s",
						logId->value.string.elements);
		break;

	default:

		/*
		 * There is a problem.
		 */
		if (cont->trace) {
		}
		free_logdb_file(log);
		return(NULL);
	}

	/*
	 * Copy the name of the file into the local log structure.
	 */
	log->fileName = (char *) malloc(strlen(filename) + 1);
	if (log->fileName == NULL) {
		if (cont->trace) {
			printf("logDB_create: out of memory.\n");
		}
		free_logdb_file(log);
		return(NULL);
	}
	strcpy(log->fileName, filename);

	/*
	 * Try to create this file.
	 */
	log->fd = fopen(filename, "w+");
	if (log->fd == NULL) {
		/*
		 * Could not create file.
		 */
		if (cont->trace) {
			printf("logDB_create: could not created file \'%s\'\n",
								filename);
		}
		free_logdb_file(log);
		return(NULL);
	}

	/*
	 * Everything OK. Log is initialized.
	 */
	if (cont->trace) {
		printf("logDB_create: end, name = \'%s\', ptr = %x\n",
						log->fileName, log);
	}
	return((logDB_handle) log);
}

/******************************************************************************/
/* Function: logDB_delete                                                     */
/*                                                                            */
/* Delete the physical log instance. Here we do the following:                */
/* 1. close the file.                                                         */
/* 2. delete (unlink) the file.                                               */
/* 3. free the memory used to hold information about the physical log.        */
/*                                                                            */
/******************************************************************************/

OM_return_code
logDB_delete(
	ca_context_t	*cont,
	logDB_handle	db
	) {

	logdb_file	*log = (logdb_file *) db;

	if (cont->trace) {
		printf("logDB_delete: start, ptr = %x\n", log);
	}

	/*
	 * Close and delete the file.
	 */
	fclose(log->fd);
	unlink(log->fileName);

	/*
	 * De-allocate the local log structure.
	 */
	free_logdb_file(log);

	if (cont->trace) {
		printf("logDB_delete: end, ptr = %x\n", log);
	}

	return(OM_SUCCESS);
}

/******************************************************************************/
/* Function: logDB_save                                                       */
/*                                                                            */
/* Save the log.  Here we do the following:                                   */
/* 1. create a name for the file to which the log will be backed up.  This    */
/*	name includes a .sav suffix and a timestamp.                          */
/* 2. create a name for the index file which will store the offsets to the    */
/* 	various log records.                                                  */
/* 3. copy the contents of the log to the backup file.                        */
/* 4. copy the list of log record information blocks to the index file.       */
/* 5. if the log is to be cleaned, delete the log file and recreate it.       */
/*                                                                            */
/******************************************************************************/

OM_return_code
logDB_save(
	ca_context_t	*cont,
	logDB_handle	db,
	OM_uint32	cleanLog
	) {

	logdb_file	*log = (logdb_file *) db;
	logdb_rec	*rec;
	OM_uint32	size;
	FILE		*fd;
	char		filename[MAXPATHLEN];
	char		filename_indx[MAXPATHLEN];
	char		*p;
	time_t		clock;
	struct tm	*tim;

	/*
	 * Below is the time format used by strftime().
	 * Note that this strange way to declare the format string 
	 * avoids side effects due to SCCS '%' subsitution
	 */
	static char timfmt[] = 
		{'%','Y','%','m','%','d','%','H','%','M','%','S','\0'};

	if (cont->trace) {
		printf("logDB_save: start, ptr = %x\n", log);
	}

	/*
	 * The save operation is to save the contents of the log file to
	 * a file with the suffix .sav, followed by a timestamp.
	 * This file can be retrieved or processed later.
	 */
	strcpy(filename, log->fileName);
	strcat(filename, ".sav.");
	p = filename + strlen(filename);
	clock = time(0);
	tim = gmtime(&clock);
	strftime(p, MAXPATHLEN - strlen(filename) - 1, timfmt, tim);

	/*
	 * Copy the file.
	 */
	if (copy_file(log->fileName, filename) != 0) {
		if (cont->trace) {
			printf("logDB_save: could not copy file \'%s\' to \'%s\' for ptr = %x\n",
						log->fileName, filename, log);
		}
		return(OM_FUNCTION_DECLINED);
	}

	/*
	 * We also need to generate a file containing the indexes to the
	 * records.  In the case of a wrapped log the file may not contain
	 * a list of contiguous records.
	 */
	strcpy(filename_indx, log->fileName);
	strcat(filename_indx, ".index.");
	p = filename_indx + strlen(filename_indx);
	strftime(p, MAXPATHLEN - strlen(filename_indx) - 1, timfmt, tim);

	/*
	 * Open the index file and write a series of records.
	 */
	if ((fd = fopen(filename_indx, "w+")) == NULL) {
		/*
		 * Cannot open file.
		 * Delete the copy and do not clean the log.
		 */
		unlink(filename);
		return(OM_FUNCTION_DECLINED);
	}

	rec = log->rec;
	size = sizeof(logdb_rec);
	while (rec != NULL) {
		if (fwrite(rec, size, 1, fd) != 1) {
			fclose(fd);
			unlink(filename_indx);
			unlink(filename);
			return(OM_FUNCTION_DECLINED);
		}
		rec = rec->next;
	}
	fclose(fd);
	

	/*
	 * Check if we are to clean the log or not.
	 */
	if (cleanLog != 0) {

		/*
		 * We need to delete the contents of the file, and then
		 * do some cleanup on the local log structure.
		 */
		if (cont->trace) {
			printf("logDB_save: cleaning log, ptr = %x\n", log);
		}

		/*
		 * Unlink the file and recreate it.
	 	 */
		unlink(log->fileName);
		log->fd = fopen(log->fileName, "w+");
		if (log->fd == NULL) {
			if (cont->trace) {
				printf("logDB_save: cannot open \'%s\', ptr = %x\n",
						log->fileName, log);
			}
			return(OM_FUNCTION_DECLINED);
		}

		/*
		 * Clean the list of records.
		 */
		free_logdb_rec(log);
		log->woff = 0;
	}

	if (cont->trace) {
		printf("logDB_save: end, ptr = %x\n", log);
	}

	return(OM_SUCCESS);
}

/******************************************************************************/
/* Function: logDB_param                                                      */
/*                                                                            */
/* Record the change in log attributes.  In this implementation, we do not    */
/* actually need to do anything when the maximum log size or log full action  */
/* attribute values change.                                                   */
/*                                                                            */
/******************************************************************************/

OM_return_code
logDB_param(
	ca_context_t	*cont,
	logDB_handle	db,
	OM_uint32	logWrap,
	OM_uint32	maxLogSize
	) {

	logdb_file	*log = (logdb_file *) db;

	if (cont->trace) {
		printf("logDB_param: start, ptr = %x, logWrap = %d, maxLogSize = %d\n",
						log, logWrap, maxLogSize);
	}

	/*
	 * Record the parameters.
	 * Note that since all the management of the deletion of records is
	 * handled by the implementation of the log.  These parameters may
	 * not be useful in all situations.
	 */
	log->logWrap = logWrap;
	log->maxLogSize = maxLogSize;

	if (cont->trace) {
		printf("logDB_param: end, ptr = %x\n", log);
	}

	return(OM_SUCCESS);
}

/******************************************************************************/
/* Function: logRecord_store                                                  */
/*                                                                            */
/* Store a log record in the log:                                             */
/* 1. store the information about the log record in an information block.     */
/* 2. copy the log record into the file at the current write offset.          */
/*                                                                            */
/******************************************************************************/

OM_return_code
logRecord_store(
	ca_context_t	*cont,
	logDB_handle	db,
	OM_uint32	id,
	ca_dmi_logR_inline_t	*recInline
	) {

	logdb_file	*log = (logdb_file *) db;
	logdb_rec	*rec;
	OM_uint32	retlen;

	if (cont->trace) {
		printf("logRecord_store: start, ptr = %x, id = %d\n", log, id);
		printInline(recInline);
		printf("\n\n");
	}

	/*
	 * Allocate some memory to hold the details of this record.
	 */
	rec = (logdb_rec *) malloc(sizeof(logdb_rec));
	if (rec == NULL) {
		if (cont->trace) {
			printf("logRecord_store: out of memory\n");
		}
		return(OM_MEMORY_INSUFFICIENT);
	}

	/*
	 * Store the details.
	 */
	rec->id = id;
	rec->size = recInline->size;
	rec->pos = log->woff;
	rec->next = NULL;
	if (log->last != NULL) {
		log->last->next = rec;
	}
	else {
		log->rec = rec;
	}
	log->last = rec;

	/*
	 * Store the record in the file.
	 */
	fseek(log->fd, log->woff, SEEK_SET);
	if ((retlen = fwrite(recInline->rec, recInline->size, 1, log->fd)) !=
							1) {
		printf("logRecord_store: error in write\n");
		return(OM_FUNCTION_DECLINED);
	}
	fflush(log->fd);

	/*
	 * Update the pointer to point to new position.
	 */
	log->woff = ftell(log->fd);

	if (cont->trace) {
		printf("logRecord_store: end, ptr = %x, id = %d\n", log, id);
	}

	return(OM_SUCCESS);
}

/******************************************************************************/
/* Function: logRecord_fetch                                                  */
/*                                                                            */
/* Get the size of a log record with a given log record ID.  Use the          */
/* information block for the log record to get this information.  There is no */
/* need to look in the file.                                                  */
/*                                                                            */
/******************************************************************************/

OM_return_code
logRecord_fetch(
	ca_context_t	*cont,
	logDB_handle	db,
	OM_uint32	id,
	OM_uint32	*size
	) {

	logdb_file	*log = (logdb_file *) db;
	logdb_rec	*rec;

	if (cont->trace) {
		printf("logRecord_fetch: start, ptr = %x, id = %d\n", log, id);
	}

	/*
	 * I need to return the size, I have this in on the record information.
	 * list.
	 */
	rec = find_logdb_rec(log, id);
	if (rec == NULL) {
		if (cont->trace) {
		}
		return(OM_FUNCTION_DECLINED);
	}
	*size = rec->size;

	if (cont->trace) {
		printf("logRecord_fetch: end, ptr = %x, id = %d, size = %d\n",
								log, id, *size);
	}

	return(OM_SUCCESS);
}

/******************************************************************************/
/* Function: logRecord_delete                                                 */
/*                                                                            */
/* Delete a log record from the file.  In fact what we do here is move the    */
/* delete offset so that this record can be overwritten if required.  The     */
/* information block is also deleted.                                         */
/*                                                                            */
/******************************************************************************/

OM_return_code
logRecord_delete(
	ca_context_t	*cont,
	logDB_handle	db,
	OM_uint32	id
	) {

	logdb_file	*log = (logdb_file *) db;
	logdb_rec	*rec;

	if (cont->trace) {
		printf("logRecord_delete: start, ptr = %x, id = %d\n", log, id);
	}

	/*
	 * Deleting a record somewhere in a file is hard to do efficiently.
	 * However, since the log implementation handles the allocation and
	 * deallocation of space for us, we can assume that old records can
	 * simply be overwritten, and therefore do not have to be explicity
	 * deleted.
	 * If the location of this record is earlier in the file than the
	 * previous one, then we can assume that the log has wrapped.
	 */
	rec = find_logdb_rec(log, id);
	if (rec->pos < log->doff || rec->pos == 0) {
		/*
		 * Log has wrapped.  Set write pointer to start of file.
		 */
		log->woff = 0;
		printf("logRecord_delete: wrap detected, ptr = %x, id = %d\n",
								log, id);
	}
	log->doff = rec->pos;

	/*
	 * Delete the information block for the record.
	 */
	delete_logdb_rec(log, id);

	if (cont->trace) {
		printf("logRecord_delete: end, ptr = %x, id = %d\n", log, id);
	}

	return(OM_SUCCESS);
}
